const body = document.createElement('body')
const root = document.createElement('div')
document.title = "Strapp.io Client"
body.appendChild(root)
document.body = body
const conf = {"iceServers": [{ "urls": "stun:stun.1.google.com:19302" }] }
let dataChannel

/* TODO: duplicate in both client.js and host.js */
function getPublicKey() {
  return new Promise( (resolve, reject) => {
    /* Check local storage for public key */
    if (!window.localStorage.getItem('public-key')) {
      /* If doesn't exist, generate public and private key pair, store in
      local storage */
      crypto.subtle.generateKey(
        { name:'RSA-OAEP',
          modulusLength: 2048,
          publicExponent: new Uint8Array([0x01, 0x00, 0x01]),
          hash: {name: "SHA-256"}
        },
        true,
        ['encrypt', 'decrypt']
        ).then((keyPair) => {
          /* TODO: Do we need to store the private key as well? */
          crypto.subtle.exportKey('jwk', keyPair.publicKey)
          .then((exportedKey) => {
            window.localStorage.setItem('publicKey', exportedKey)
            console.log('public key is' + window.localStorage.getItem('publicKey'))
            resolve(exportedKey)
          })

        })
    }
    else {
      resolve(window.localStorage.getItem('publicKey'))
    }
  })

}

function postServer(url, data) {
  const request = new XMLHttpRequest()
  request.open('POST', url, true)
  request.setRequestHeader('Content-Type', 'application/json' )
  request.setRequestHeader('X-Strapp-Type', 'ice-candidate-submission')
  request.send(data)
}

/* TODO: All this does is wrap a function in a promise. Allows pollServerForAnswer
         to call itself recursively with the same promise */
function pollServer(url, clientPubKey, func) {
  return new Promise((resolve, reject) => {
    func(url, clientPubKey, resolve, reject )
  })
}

/* Poll the server. Send get request, wait for timeout, send another request.
   Do this until...? Can be used for either reconnecting or waiting for answer*/
function pollServerForAnswer(url, data, resolve, reject) {
  const request = new XMLHttpRequest()
  request.open('GET', url, true)
  /* But there is no JSON? */
  request.setRequestHeader('Content-Type', 'application/json' )
  request.setRequestHeader('X-Strapp-Type', 'client-sdp-offer')
  request.setRequestHeader('X-Client-Offer', JSON.stringify(data))
  request.onreadystatechange = () => {
    if (request.status === 200) {
      if(request.readyState === 4) {
        console.log('Client: Recieved Answer from Host')
        console.log(request)
        resolve(request.response)
      }
    }
    else if (request.status === 504) {
      console.log('timed out, resending')
      pollServerForAnswer(url, data, resolve, reject)
    }
    else {
      reject('server unhandled response of status ' + request.status)
    }
  }
  request.send()
}

/* Poll server for ice candidates until ice is complete */
function pollServerForICECandidate(cpc, url, pubKey) {
  let intervalID = window.setInterval(() => {
    if (cpc.iceConnectionState.localeCompare('connected') !== 0
      && cpc.iceConnectionState.localeCompare('completed') !== 0) {
        console.log('Client: Polling server begin for intervalID = ' + intervalID)
         console.log('Client: Requesting ICE Candidates from server')
         const request = new XMLHttpRequest()
         request.open('GET', url, true)
         request.setRequestHeader('Content-Type', 'application/json' )
         request.setRequestHeader('X-Strapp-Type', 'ice-candidate-request')
         request.setRequestHeader('X-client-pubkey', pubKey)
         request.onreadystatechange = () => {
           if (request.status === 200) {
             if(request.readyState === 4) {
               console.log('Client: Recieved ICE Candidate from Host')
               cpc.addIceCandidate(new RTCIceCandidate(JSON.parse(request.response).ice))
             }
           }
           else if (request.status === 204) {
             console.log('Ice Candidate unavailable, trying again in one second')
           }
           else {
             console.log('server unhandled response of status ' + request.status)
             clearInterval(intervalID)
           }
         }
         request.send()
    }
    else {
      clearTimeout()
      clearInterval(intervalID)
    }
  }, 5000)
}

/* Create and send offer -> Send ICE Candidates -> Poll for ICE Candidates */
getPublicKey().then((cpk) => {
  console.log('Client: Create and send offer')
  const cpc = new RTCPeerConnection(conf)

  cpc.oniceconnectionstatechange = () => {
    console.log('iceConnectionState = ' + cpc.iceConnectionState)
  }

  cpc.onnegotiationneeded = () => {
    console.log('negotiation needed!')
    cpc.createOffer().then((offer) => {
      return cpc.setLocalDescription(offer)
    })
    .then(() => {
      console.log('Client: Sending offer to host')
      let offer = {
        cmd: '> sdp pubKey',
        sdp: cpc.localDescription,
        pubKey: cpk.n
      }
      return pollServer(window.location, offer, pollServerForAnswer)
    }).then((serverResponse) => {
      const answer = JSON.parse(serverResponse)
      console.log('Client: Polling for ICE candidates')
      pollServerForICECandidate(cpc, window.location, cpk.n)
      cpc.setRemoteDescription(answer.sdp)
      cpc.onicecandidate = (event) => {
        if (event.candidate) {
          console.log('Client: Sending ice candidate to host')
          postServer(window.location, JSON.stringify({
            cmd: '> ice pubkey',
            ice: event.candidate,
            pubKey: cpk.n
          }))
        }
        else {
          console.log('Client: No more Ice Candidates to send')
        }
      }


    }).catch( (err) => {
      console.log('error in sdp handshake: ' + err)
    })
  }
  /* Start data channel */
  dataChannel = cpc.createDataChannel("sendChannel");
  dataChannel.onmessage = (msg) => {
    console.log(msg.data)
  }
  dataChannel.onopen = () => {
    dataChannel.send(`Hi from the Client`)
  }

})
